# Copyright (c) 2019 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from pako.config_loader import recursive_merge, load_package_managers_overrides

__package_managers = {
    'eopkg': {
        'sudo': True,
        'update': 'ur',
        'install': 'it -y',
        'formats': {
            'exe': ['{}', '{}-utils', '{}-bin'],
            'lib': ['{}', 'lib{}'],
            'lib-dev': ['{}-devel', 'lib{}-devel'],
            'lib-debug': ['{}-dbginfo', 'lib{}-dbginfo'],
        }
    },
    'apt-get': {
        'sudo': True,
        'update': 'update',
        'install': 'install',
        'formats': {
            'exe': ['{}', '{}-utils'],
            'lib': ['lib{}', '{}'],
            'lib-dev': ['lib{}-dev', '{}-dev'],
            'lib-debug': ['lib{}-dbg', '{}-dbg']
        }
    }
}


def get_package_manager_names():
    return list(__package_managers)


def load_package_manager_data():
    return dict(recursive_merge(__package_managers, load_package_managers_overrides()))



# Copyright (c) 2019 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import appdirs
import json
from logging import getLogger
from os import makedirs
from os.path import join, isfile


def recursive_merge(a, b):
    """Returns a generator for the merged dict of a and b"""
    for k in set(a.keys()) | set(b.keys()):
        if k in a and k in b:
            if isinstance(a[k], dict) and isinstance(b[k], dict):
                yield k, dict(recursive_merge(a[k], b[k]))
            else:
                yield k, b[k]
        elif k in a:
            yield k, a[k]
        else:
            yield k, b[k]


def load_package_managers_overrides() -> dict:
    """Load the optional package manager data overrides from disk"""
    pako_config = appdirs.user_config_dir('pako')
    config_file = join(pako_config, 'pako.conf')
    makedirs(pako_config, exist_ok=True)

    if not isfile(config_file):
        from pako.package_manager_data import get_package_manager_names
        with open(config_file, 'w') as f:
            json.dump({i: {} for i in get_package_manager_names()}, f)
    try:
        with open(config_file) as f:
            custom_config = json.load(f)
    except ValueError:
        getLogger(__name__).warning('Failed to load config file')
        custom_config = {}

    return custom_config



# Copyright (c) 2019 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Tuple


class PackageFormat:
    """
    Possible categories a package can be classified as
    This is used to guess name variants in different package managers
    """
    exe = 'exe'
    lib = 'lib'
    lib_dev = 'lib-dev'
    lib_debug = 'lib-debug'
    none = 'none'  # no special formatting

    all = [exe, lib, lib_dev, lib_debug, none]

    @classmethod
    def parse(cls, package_string: str) -> Tuple[str, str]:
        """
        Convert package name to base package name and package format string
        Example: 'libgdb-dev' -> ('gdb', 'lib-dev')
        """

        def try_parse(param, from_end=True):
            nonlocal package_string
            if from_end:
                if package_string.endswith(param):
                    package_string = package_string[:-len(param)]
                    return True
            else:
                if package_string.startswith(param):
                    package_string = package_string[len(param):]
                    return True

        has_lib = try_parse('lib', False)
        has_dev = (
                try_parse('-dev') or
                try_parse('-devel')
        )
        has_debug = (
                try_parse('-dbg') or
                try_parse('-debug') or
                try_parse('-dbginfo')
        )
        has_exe = (
                try_parse('-utils') or
                try_parse('-bin')
        )
        if has_dev:
            fmt = cls.lib_dev
        elif has_debug:
            fmt = cls.lib_debug
        elif has_lib:
            fmt = cls.lib
        elif has_exe:
            fmt = cls.exe
        else:
            fmt = cls.exe
        return package_string, fmt



# Copyright (c) 2019 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import sys

from os.path import basename
from shutil import which
from subprocess import call, PIPE

from pako.package_format import PackageFormat
from pako.package_manager_data import load_package_manager_data


class PakoManager:
    def __init__(self):
        self.has_sudo = False
        self.all_data = load_package_manager_data()
        self.name = self._find_package_manager(list(self.all_data))

        if not self.name:
            raise RuntimeError('Package manager not found!')

        self.exe = which(self.name)
        self.config = self.all_data[self.name]
        if self.config['sudo'] and not sys.stdout.isatty() and not self._check_for_sudo():
            raise RuntimeError('Sudo required (and script not launched interactively)')

    def _check_for_sudo(self):
        return call(['sudo', '-n', 'true'], stderr=PIPE) == 0

    @staticmethod
    def _find_package_manager(exes):
        for exe in exes:
            if which(exe):
                return exe
        return None

    def call(self, args: list):
        sudo_args = []
        if self.config['sudo']:
            sudo_args = ['sudo']
            if not self.has_sudo:
                if self._check_for_sudo():
                    self.has_sudo = True
                else:
                    print('Requesting sudo to run command: {}...'.format(
                        ' '.join([basename(self.exe)] + args)
                    ))
        return call(sudo_args + [self.exe] + args)

    def update(self):
        return self.call(self.config['update'].split()) == 0

    def install_one(self, package: str, fmt: str = None) -> bool:
        if not fmt:
            package, fmt = PackageFormat.parse(package)
        if fmt not in PackageFormat.all:
            raise ValueError('Invalid package format: {}. Should be one of: {}'.format(
                fmt, PackageFormat.all
            ))

        possible_names = []
        for format_name, formats in self.config['formats'].items():
            if not fmt or format_name == fmt:
                for f in formats:
                    if f not in possible_names:
                        possible_names.append(f)

        for name in possible_names:
            if self.call(self.config['install'].split() + [name.format(package)]) == 0:
                return True
        return False

    def install(self, packages: list, overrides: dict = None) -> bool:
        if isinstance(packages, str):  # Easy mistake
            raise TypeError('Packages parameter must be a list')
        overrides = overrides or {}
        if self.name in overrides:
            packages = overrides[self.name]
            return self.call(self.config['install'].split() + packages) == 0
        for package in packages:
            if not self.install_one(package):
                return False
        return True



# Copyright (c) 2019 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from setuptools import setup

setup(
    name='pako',
    version='0.2.0',
    packages=['pako'],
    install_requires=['appdirs'],
    url='https://github.com/MycroftAI/pako',
    license='Apache-2.0',
    author='Mycroft AI',
    author_email='support@mycroft.ai',
    maintainer='Matthew Scholefield',
    maintainer_email='matthew331199@gmail.com',
    description='The universal package manager library',
    entry_points={
        'console_scripts': {
            'pako=pako.__main__:main'
        }
    },
    data_files=[('pako', ['LICENSE'])]
)



# Copyright (c) 2019 Mycroft AI, Inc.
#
# This file is part of Mycroft Skills Manager
# (see https://github.com/MatthewScholefield/mycroft-light).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from argparse import ArgumentParser

from pako.pako_manager import PakoManager


def install(args):
    return PakoManager().install(args.packages)


def update(args):
    return PakoManager().update()


def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest='action')
    p = subparsers.add_parser('install')
    p.add_argument('packages', nargs='+')
    subparsers.add_parser('update')
    args = parser.parse_args()
    try:
        result = {'install': install, 'update': update}[args.action](args)
        if result:
            print('{} succeeded'.format(args.action.title()))
            exit(0)
        else:
            print('{} failed'.format(args.action.title()))
            exit(1)
    except KeyboardInterrupt:
        print('Cancelled.')
    except Exception as e:
        print('Error: ' + str(e))
